package java.util;

import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import sun.misc.SharedSecrets;
import java.io.Serializable;

public abstract class EnumSet<E extends Enum<E>> extends AbstractSet<E> implements Cloneable, Serializable
{
    final Class<E> elementType;
    final Enum<?>[] universe;
    private static Enum<?>[] ZERO_LENGTH_ENUM_ARRAY;
    
    EnumSet(Class<E> elementType, Enum<?>[] universe) {
        this.elementType = elementType;
        this.universe = universe;
    }
    
    public static <E extends Enum<E>> EnumSet<E> noneOf(Class<E> obj) {
        Enum[] universe = getUniverse(obj);
        if (universe == null) {
            throw new ClassCastException(obj + " not an enum");
        }
        if (universe.length <= 64) {
            return new RegularEnumSet(obj, universe);
        }
        return new JumboEnumSet(obj, universe);
    }
    
    public static <E extends Enum<E>> EnumSet<E> allOf(Class<E> clazz) {
        EnumSet<Enum> none = noneOf(clazz);
        none.addAll();
        return none;
    }
    
    abstract void addAll();
    
    public static <E extends Enum<E>> EnumSet<E> copyOf(EnumSet<E> set) {
        return set.clone();
    }
    
    public static <E extends Enum<E>> EnumSet<E> copyOf(Collection<E> collection) {
        if ((collection instanceof EnumSet)) {
            return ((EnumSet)collection).clone();
        }
        if (collection.isEmpty()) {
            throw new IllegalArgumentException("Collection is empty");
        }
        Iterator<Object> iterator = collection.iterator();
        EnumSet<Object> of = of((Enum)iterator.next());
        while (iterator.hasNext()) {
            of.add(iterator.next());
        }
        return of;
    }
    
    public static <E extends Enum<E>> EnumSet<E> complementOf(EnumSet<E> set) {
        EnumSet<Enum> copy = copyOf(set);
        copy.complement();
        return copy;
    }
    
    public static <E extends Enum<E>> EnumSet<E> of(E e) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        return none;
    }
    
    public static <E extends Enum<E>> EnumSet<E> of(E e, E e2) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        none.add(e2);
        return none;
    }
    
    public static <E extends Enum<E>> EnumSet<E> of(E e, E e2, E e3) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        none.add(e2);
        none.add(e3);
        return none;
    }
    
    public static <E extends Enum<E>> EnumSet<E> of(E e, E e2, E e3, E e4) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        none.add(e2);
        none.add(e3);
        none.add(e4);
        return none;
    }
    
    public static <E extends Enum<E>> EnumSet<E> of(E e, E e2, E e3, E e4, E e5) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        none.add(e2);
        none.add(e3);
        none.add(e4);
        none.add(e5);
        return none;
    }
    
    @SafeVarargs
    public static <E extends Enum<E>> EnumSet<E> of(E e, E... array) {
        EnumSet none = noneOf(e.getDeclaringClass());
        none.add(e);
        for (int length = array.length, i = 0; i < length; ++i) {
            none.add(array[i]);
        }
        return none;
    }
    
    public static <E extends Enum<E>> EnumSet<E> range(E obj, E e) {
        if (obj.compareTo(e) > 0) {
            throw new IllegalArgumentException(obj + " > " + e);
        }
        EnumSet none = noneOf(obj.getDeclaringClass());
        none.addRange(obj, e);
        return none;
    }
    
    abstract void addRange(E p0, E p1);
    
    public EnumSet<E> clone() {
        try {
            return (EnumSet)super.clone();
        }
        catch (CloneNotSupportedException detailMessage) {
            throw new AssertionError(detailMessage);
        }
    }
    
    abstract void complement();
    
    final void typeCheck(E e) {
        Class<Object> class1 = e.getClass();
        if (class1 != this.elementType && class1.getSuperclass() != this.elementType) {
            throw new ClassCastException(class1 + " != " + this.elementType);
        }
        return;
    }
    
    private static <E extends Enum<E>> E[] getUniverse(Class<E> clazz) {
        return SharedSecrets.getJavaLangAccess().getEnumConstantsShared(clazz);
    }
    
    Object writeReplace() {
        return new EnumSet.SerializationProxy(this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }
    
    static {
        EnumSet.ZERO_LENGTH_ENUM_ARRAY = new Enum[0];
    }
    
    private static class SerializationProxy<E extends Enum<E>> implements Serializable
    {
        private final Class<E> elementType;
        private final Enum<?>[] elements;
        private static final long serialVersionUID = 362491234563181265L;
        
        SerializationProxy(EnumSet<E> set) {
            this.elementType = set.elementType;
            this.elements = (Enum[])set.toArray(EnumSet.ZERO_LENGTH_ENUM_ARRAY);
        }
        
        private Object readResolve() {
            EnumSet<E> none = EnumSet.noneOf(this.elementType);
            Enum[] elements = this.elements;
            for (int length = elements.length, i = 0; i < length; ++i) {
                none.add(elements[i]);
            }
            return none;
        }
    }
}
